;
; dos.asm - Common dos functions
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;---------------------------------------
; hdinit
; C=1 if error then drive_type_exp is 00
;
; fills in drive_type_exp with 01/80/c0
;--------------------------------------
OP_hdinit	.proc
		sta current_drive_exp

		jsr OP_select_common	;Select drive

		jsr defaultwait
lp2     	jsr wait_a_bit
		jsr wait_a_bit
		jsr BACK_with_C1_on_timeout_or_C0
		lda status_g
		lda low_reg
                cmp #$ff
                beq nosig
                asl
		bcs lp2                 ;Wait BSY==0 and DRDY=1

                jsr wait_50ms

                lda error_g
                lda low_reg
                cmp #$01		;Check diagnostic code
                beq kq
                cmp #$81
                bne nosig		;not valid
kq
		lda cyllow_g		;check ATA/ATAPI signature
		lda low_reg
		bne no_ata
		lda cylhi_g
		lda low_reg
		bne nosig

		lda status_g
		lda low_reg
                beq ok			;probably nothing here

		ldy #0
lp1     	dex
		bne jx
		dey
		beq ok
jx		jsr wait_a_bit
		lda status_g
		lda low_reg
                asl
                bpl lp1			;wait DRDY, but not forever
ok
		lda #ident_d
		jsr identify_drive
		bcs nosig

yes_hdd		lda #$01		;*yes* drive=HDD
		.byte SKIP2
yes_cd		lda #$80		;*yes* drive=CDROM
		.byte SKIP2
yes_floppy	lda #$C0		;*yes* drive=ZIP

		#ldx_l_ current_drive_exp
		sta drive_type_exp,x
                lda ident_maxhead_exp
                #sbc1c
                sta drive_maxhead_exp,x
                lda ident_maxsector_exp
                sta drive_maxsector_exp,x
		clc
		rts

nosig		sec
		rts

no_ata		eor #$14
		bne nosig		;check atapi signature
		lda cylhi_g
		lda low_reg
		eor #$eb
		bne nosig
		lda #atapi_reset_d	
		jsr ata_command
                bcs nosig
		#stz_a low_reg
		sta feat_g		;for some atapi drives
		lda #atapi_ident_d	
		jsr identify_drive
		bcs nosig

		lda ident_config_exp
		and #31
		cmp #5			;CDROM?
		beq yes_cd
					;Floppy/Unknown, but it may work (if sector size=512)
		ldy #249
		ldx #215
j11		iny
		beq yes_cd		;Looks like an old CD-ROM
j1		inx
		beq yes_floppy		;It's a floppy
		lda ident_model_number_exp-216,x
		cmp cdtext-250,y
		beq j11
cc		cpy #250
		beq j1
		dex
		dey
		gne cc

cdtext		.byte $43,$44,$2d,$52,$4f,$4d
		.pend
;---------------------------------------
; Identify Drive
; C=1 if error
;
; reads indentify info from drive
;---------------------------------------
identify_drive	.proc
		jsr ata_data_command
		bcs identerr		;error C=1

		lda #ram_exp_buffer0
		jsr OP_read512_here

		ldx #182		;fix ascii strings
cp		lda ident_serial_number_exp-182,x
		pha
		lda ident_serial_number_exp-181,x
		sta ident_serial_number_exp-182,x
		pla
		sta ident_serial_number_exp-181,x
		inx
		inx
		bne cp
		geq ata_pi_nomore_data
identerr	rts
		.pend
;---------------------------------------
; misc settings for existing drives
;
; Also sets the drive identify message
;--------------------------------------
OP_XIT_hdinit_2	.proc
		sta current_drive_exp
		tax
		lda drive_type_exp,x
		beq no

		#xit_init

		jsr set_power_management_and_features
		ldx #0
		ldy #0
		jmp LWOP_reporterror_73_xit
no		rts
		.pend
;--------------------------------------
; Tell an ata drive the disk geometry
;
OP_set_hdd_max	.proc
		jsr OP_get_current_drive_type
		bmi err			;for hdd only

		lda drive_maxhead_exp,x
		tay
		asl
		bmi err			;for chs only
		lda drive_maxsector_exp,x
		pha
		jsr ata_select_drive
		pla
		bcs err			;error C=1
		sta low_reg
		sta nsector_g
		lda #init_d
		gne ata_command
err		rts
		.pend
;--------------------------------------
ofs_fs_ident	.text "C64-IDE V0.02-A "
cfs_fs_ident	.text "C64 CFS V 0.11B "
;--------------------------------------
; Try to identify filesystem on disk
;
OP_XIT_init_filesystem_from_disk .proc

		jsr OP_get_current_drive_type
		and #<~FS_MASK
		sta drive_type_exp,x	;Unknown filesystem
		bpl ata
		asl
                php

		lda #ATAPI_TEST_UNIT_READY
		jsr OP_atapi_packet
		jsr OP_atapi_command	;double ATAPI_SENSE_UNIT_ATTENTION after reset

                plp
		bmi zip
                
                #stz_a current_path

                lda #ATAPI_READ_TOC
                jsr OP_atapi_packet
                #pc 9,$40
                #pc 7,>$800
                jsr OP_atapi_datain_command
                bcs unsupp
                lda #ram_exp_buffer0
		jsr OP_read512_here
		jsr OP_read512
		jsr OP_read512
		jsr OP_read512
		jsr ata_pi_nomore_data
                bcs unsupp
                lda toc_first_session_number_exp
                eor toc_last_session_number_exp
                beq unsupp
                ldx #3
                lda #15
                .if SUPERCPU=1
                .byte SKIP1
lu              tdc
                .else
                .byte SKIP2
lu              lda #0
		.fi
                adc toc_last_session_lba_exp,x
                sta next_sector,x
                dex
                bpl lu
                gmi xi

unsupp          lda #15
		sta next_sector+3
		#stz_a next_sector
		#stz_aa next_sector+1
		#stz_aa next_sector+2

xi              inc current_path
		bmi ene
                ldx #3
ll      	inc next_sector,x
                bne xi2
                dex
		bpl ll
ene		jsr OP_get_current_drive_type
		and #FS_MASK
		bne LWOP_reporterror_ok
                sec
                bcs LWOP_reporterror_75_xit;filesystem not found

xi2     	lda #ram_exp_buffer0
		jsr XIT_read_a_sector2
                
                lda ram_exp_buffer0_exp
                cmp #1
                beq bp
                cmp #2
                beq bp
                cmp #255
                bne xi
bp              tay

                ldx #4
ly              lda cdstr,x
                cmp ram_exp_buffer0_exp+1,x
                bne xi
                dex
                bpl ly
                iny
                beq ene		;0xff = end

		#ldatax current_drive_exp
		#lda0			;First partition is active
		sta drive_workpartition_exp,x
                txa
		lsr
		ror			;*64
		ror
		tax

		cpy #3
                php
		lda #%10000000		;Make partition valid, not hidden and readonly
		sta partition_start_exp,x
		lda #%00010000
		sta partition_end_exp,x

                ldy #3
                sec
luk             #lda_l_y_ ram_exp_buffer0_exp+158
		bcc *+5
                ora #$40
		clc
		sta partition_root_dir_exp,x
		sta partition_workdir_exp,x
                inx
                dey
                bpl luk

                jsr inval

		jsr OP_get_current_drive_type
                plp
                beq jol
		ora #FS_ISO
		sta drive_type_exp,x	;ISO filesystem
                gne xi			;retry
                
jol		ora #FS_JOLIET
		sta drive_type_exp,x	;Joliet filesystem
                gne LWOP_reporterror_ok

cdstr           .text "CD001"

zip		#lda0			;ZIP drive, LBA sector 0
		.byte SKIP2
ata		lda #1			;ATA CHS (0/0/1)
		sta next_sector+3
		#stz_a next_sector
		#stz_aa next_sector+1
		#stz_aa next_sector+2

		lda #ram_exp_buffer0
		jsr XIT_read_a_sector2

		ldx #15
lp1		lda boot_sector_ident_cfs_exp,x
		eor cfs_fs_ident,x
		bne inifi_nocfs
		dex
		bpl lp1

		#ldatax current_drive_exp;Store disk geometry
		lda boot_sector_maxhead_cfs_exp
		and #%01001111
		sta drive_maxhead_exp,x
		lda boot_sector_maxcyllow_cfs_exp
		sta drive_maxcylinderlo_exp,x
		lda boot_sector_maxcylhi_cfs_exp
		sta drive_maxcylinderhi_exp,x
		lda boot_sector_maxsector_cfs_exp
		sta drive_maxsector_exp,x

		lda boot_sector_default_p_cfs_exp
		sta drive_workpartition_exp,x

		jsr OP_set_hdd_max	;Tell disk the new geometry

		#cnext boot_sector_partdir1_cfs_exp , next_sector

		lda #ram_exp_buffer0
		jsr XIT_read_a_sector2	;-=> error

		lda #ram_exp_buffer0
		jsr init_pom1_y
		lda current_drive_exp
		lsr
		ror			;*64
		ror
		tax

loop		ldy #16
lp2		#lda_zp_y pom1
		sta partition_start_exp,x
		iny
		iny
		iny
		iny
		#lda_zp_y pom1
		sta partition_end_exp,x
		iny
		iny
		iny
		iny
		#lda_zp_y pom1
		sta partition_del_dir_exp,x
		iny
		iny
		iny
		iny
		#lda_zp_y pom1
		sta partition_root_dir_exp,x
		sta partition_workdir_exp,x
		inx
		tya
		sec
		sbc #11
		tay
		and #3
		bne lp2
		lda pom1
		adc #32-1		;C=1
		sta pom1
		bne loop
		inc pom2
		txa
		and #63
		bne loop
		#lda0
		jsr OP_check_partition_selection
                jsr LWOP_reporterror_77_xit;Invalid? -=> error
		jsr OP_get_current_drive_type
		ora #FS_CFS
		sta drive_type_exp,x	;CFS filesystem
		jmp LWOP_reporterror_ok

;----------------------------
inifi_nocfs
		ldx #15
lp1c		lda boot_sector_ident_ofs_exp,x
		eor ofs_fs_ident,x
		bne inifi_noofs
		dex
		bpl lp1c

		#ldatax current_drive_exp
		lda boot_sector_maxhead_ofs_exp
		and #%00001111		;no lba here!
		sta drive_maxhead_exp,x
		lda boot_sector_maxcyllow_ofs_exp
		sta drive_maxcylinderlo_exp,x
		lda boot_sector_maxcylhi_ofs_exp
		sta drive_maxcylinderhi_exp,x
		lda boot_sector_maxsector_ofs_exp
		sta drive_maxsector_exp,x

		#lda0			;First partition is active
		sta drive_workpartition_exp,x

		txa
		lsr
		ror			;*64
		ror
		tax

		lda #%10000000		;Make partition valid, not hidden and readonly
		sta partition_start_exp,x
		lda #%00010000
		sta partition_end_exp,x

		lda boot_sector_root_ofs_exp+2
		and #15			;setup root and workdir
		sta partition_root_dir_exp,x
		sta partition_workdir_exp,x
		inx
		lda boot_sector_root_ofs_exp
		sta partition_root_dir_exp,x
		sta partition_workdir_exp,x
		inx
		lda boot_sector_root_ofs_exp+1
		sta partition_root_dir_exp,x
		sta partition_workdir_exp,x
		inx
		lda boot_sector_root_ofs_exp+3
		sta partition_root_dir_exp,x
		sta partition_workdir_exp,x
		inx

		jsr inval

		jsr OP_set_hdd_max

		jsr OP_get_current_drive_type
		ora #FS_OFS
		sta drive_type_exp,x	;CFS filesystem
		jmp LWOP_reporterror_ok
                
inval		#lda0			;Invalidate partitions
		sta partition_start_exp,x
		inx
                txa
		and #63
		bne inval
		rts
;----------------------------
;jotyp		.byte $1,$4,$6,$b,$c,$e,$f
inifi_noofs
		.comment
		jsr check_s
                bne inifi_nopc		;invalid

                ldy #<boot_sector_type_exp
nu              #lda_zp_y pom1
                ldx #6
lo              cmp jotyp,x
                beq fel
                dex
                bpl lo

                lda pom1
                clc
                adc #16			;search for partition entry
                sta pom1
		asl
                bpl nu
                gmi inifi_nopc		;none found

fel		ldx #3
                ldy #<(boot_sector_startp_exp+3)
re              #lda_zp_y pom1
		sta szam,x
                sta ram_exp_bitmap_exp,x  	;little endian partition start
                dey
                dex 
                bpl re

                jsr lba2chs

		lda #ram_exp_buffer0
		jsr XIT_read_a_sector2	;read boot sector

		jsr check_s
                bne inifi_nopc		;invalid

		#ldatax current_drive_exp
		#lda0			;First partition is active
		sta drive_workpartition_exp,x
		txa
		lsr
		ror			;*64
		ror
		tax
                
                #ldy_l_a boot_sector_fat_rootcount_exp+1
                lda boot_sector_fat_rootcount_exp
                and #$f0
                ora #$08

kbe             pha
                tya
                lsr
                tay			;divide by 32
                pla
                ror
                bcc kbe
                sta partition_fat_rootsect_exp,x;root sector count
                tya
                sta partition_fat_rootsect_exp+1,x

		lda boot_sector_fat_secperclust_exp
                sta partition_fat_sectperclust_exp,x;sector per cluster count

		lda #%00010000          ;not hidden and readonly
		sta partition_end_exp,x

		lda boot_sector_fat_numfat_exp
                sta next_sector

                #lda0
                sta boot_sector_fat_reserved_exp+2
                sta boot_sector_fat_reserved_exp+3

                #phx_a

		ldx #0
                clc
ru              lda boot_sector_fat_reserved_exp,x
                adc ram_exp_bitmap_exp,x	;partition start
                sta szam,x
                inx
                txa
                and #3
                bne ru

		pla
                pha
                tax

                lda partition_fat_fat_exp,x
                ora #$80
                sta partition_fat_fat_exp,x

                lda partition_fat_rootsect_exp,x
                clc
                adc szam
                sta szam
                lda partition_fat_rootsect_exp+1,x
                adc szam+1
                sta szam+1
                bcc ree2
                inc szam+2
                bne ree2
                inc szam+3


ree2		ldx #boot_sector_fat_fat2sectors_exp-boot_sector_fat_fatsectors_exp
		lda boot_sector_fat_fatsectors_exp
                ora boot_sector_fat_fatsectors_exp+1
                beq xj
                #lda0
                tax
                sta boot_sector_fat_fatsectors_exp+2
                sta boot_sector_fat_fatsectors_exp+3
xj

                ldx #0
                clc
ree             lda boot_sector_fat_fatsectors_exp,x
                adc szam,x
                sta szam,x      		;calculate data area (big endian)
		inx
                txa
                and #3
                bne ree
		dec next_sector
		bne ree2

		pla
                pha
                tax

                ldy #3
mf              lda szam,y
                sta partition_fat_data_exp,x	;data start
                inx
                dey
                bpl mf
                
                #plx_a

		#lda0
                ldy #4
jqj             sta partition_workdir_exp,x
                sta partition_root_dir_exp,x
                inx
                dey
                bne jqj

lp2cc   	#lda0			;Invalidate partitions
		sta partition_start_exp,x
		inx
                txa
		and #63
		bne lp2cc

		jsr OP_set_hdd_max

		jsr OP_get_current_drive_type
		ora #FS_FAT
		sta drive_type_exp,x	;CFS filesystem
		jmp LWOP_reporterror_ok
                .endc
;----------------------------
inifi_nopc     	sec
		gcs LWOP_reporterror_75_xit;Unknown filesystem
                
check_s	      	lda #>boot_sector_sign_exp
                sta pom2
                ldy #0
                sty pom1
                .if SUPERCPU=1
                iny
                sty pom3
                .fi
                ldy #<boot_sector_sign_exp

                lda #$55
np              #cmp_zp_y pom1
                bne nke
                eor #$ff
                iny
                bne np
nke		rts
		.pend
;----------------------------
;C=1 on error, result in Y
;
OP_check_partition_selection .proc
		sec
		sbc #1
		bcs fe
		#ldatax current_drive_exp
		lda drive_workpartition_exp,x

fe		tay
		cpy #16
		bcs invalid
		lsr
		ror
		ror
		ror
		ora current_drive_exp
		ror
		ror
		ror
		tax
		lda partition_start_exp,x
		bpl invalid		;Must be valid
		lda partition_end_exp,x
		and #%10110000
		eor #%00010000		;CFS?
		beq valid
invalid		iny
		tya
		tax
		ldy #0
                sec
                .byte SKIP1
valid		clc
		rts

		.pend
;--------------------------------------
kill_drives	.proc
		#lda0
		ldx #DRIVES-1
lp		sta drive_type_exp,x
		dex
		bpl lp
		rts
		.pend
;---------------------------------------
set_power_management_and_features .proc
		jsr head_select_drive

		lda setup_powerman_exp
		jsr OP_master_slave_drive
		bcc master
		lsr
		lsr
		lsr
		lsr
master		and #%00001111
		beq def
		tax
		lda powerconst-1,x
		sta low_reg
		sta nsector_g

		lda #idle1_d		;Power management feature
		jsr ata_command

		lda #idle2_d		;Power management feature
		jsr ata_command

def		ldx #0                  ;retry
		jsr set_features

                ldx #3                  ;writecache
		jsr set_features

                ldx #6                  ;readahead
		jsr set_features

		jsr OP_get_current_drive_type
                bpl exit
                asl
		bmi exit		;not cdrom

		lda #ATAPI_SET_SPEED
		jsr OP_atapi_packet

		lda #%00000001		;cdromspeed
		jsr OP_master_slave_drive
		bcc *+4
		lda #%00010000
		ldx #$ff		;optimal speed ($ffff)
		and setup_features_exp
		beq ns
		inx			;lowspeed ($ff)
ns
		lda #255		;255Kb/s should be enough for everyone
		#pw 3
		txa
		#pw 2
		jsr OP_atapi_command
		bcc jo
		jsr OP_atapi_command	;retry
jo
		jsr cdrom_idleoff
		
exit		rts

featc		.byte %10000000,$33,$99		;retry
		.byte %01000000,$82,$02		;writecache
                .byte %00100000,$55,$aa         ;readahead

set_features	jsr OP_master_slave_drive
		lda featc,x
		bcs nm
		lsr
                lsr
                lsr
                lsr
nm		and setup_features_exp
		beq dis
                inx
dis		lda featc+1,x
		sta low_reg
		sta feat_g
		lda #setfeatures_d
		gne ata_command

powerconst	.byte 0,12,24,36,48,60,72,84,96,120,180,240,241,242,244
		.pend
